/* qcam.c -- routines for accessing the Connectix QuickCam */

/* Version 0.4b, May 21, 1996 */
/* Version 0.5, August 24, 1996 */


/******************************************************************

Copyright (C) 1996 by Scott Laird

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL SCOTT LAIRD BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include "qcam.h"

void usage(void)
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"  qcam [options]\n");
  fprintf(stderr,"    Options:\n");
  fprintf(stderr,"      -x width   Set width\n");
  fprintf(stderr,"      -y height  Set height\n");
  fprintf(stderr,"      -p port    Set port\n");
  fprintf(stderr,"      -B bpp     Set bits per pixel\n");

  fprintf(stderr,"      -c val     Set contrast\n");
  fprintf(stderr,"      -H val     Set hue\n");
  fprintf(stderr,"      -S val     Set saturation\n");
  fprintf(stderr,"      -d val     Set black level\n");
  fprintf(stderr,"      -w val     Set white balance\n");
  fprintf(stderr,"      -W         Auto-set white balance\n");
  fprintf(stderr,"      -b val     Set brightness\n");
  fprintf(stderr,"      -E         Auto-Exposure\n");
  fprintf(stderr,"      -s val     Set scaling factor (1, 2, or 4)\n");
  fprintf(stderr,"      -t val     Set top line of scan\n");
  fprintf(stderr,"      -l val     Set left column of scan\n");
  fprintf(stderr,"      -f file    Select configuration file\n");
  fprintf(stderr,"      -u         Force unidirectional mode\n");
  fprintf(stderr,"      -V         Show version information\n");

}


void modegripe(struct qcam *q)
{
  fprintf(stderr,"Unsupported resolution/bit depth!\n");
  fprintf(stderr,"This program only supports 320x240, 160x120, and 80x60.\n");
  fprintf(stderr,"You specified %d x %d.  Try again.\n",q->width,q->height);
  exit(1);
}

/* Calculate average pixel value for entire image */

int pixel_avg(struct qcam *q, scanbuf *scan)
{
  int count = 0;
  int sum = 0;
  int pixels;
  int i;

  pixels = q->height / q->transfer_scale;
  pixels *= q->width / q->transfer_scale;

  if (q->cam_version == 0x10) {
    for (i = 0; i < (pixels * 3); i++) {
      sum += scan[i];
      count++;
    }
  } else {  
    for (i = 0; i < pixels; i++) {
      sum += scan[i];
      count++;
    }
  }
  return (sum / count);
}

/* Return false(0) if exposure is correct, otherwise adjust the
 * brightness and return true(1).
 */

int adjust_exposure(struct qcam *q, scanbuf *scan)
{
  int luminance_target;
  int luminance_dif;
  int luminance_avg;
  int brightness_cur;
  int brightness_new;
  int brightness_adj;

  switch(qc_getbitdepth(q)) {
    case 4:  luminance_target =   8;break;
    case 6:  luminance_target =  32;break;
    case 24: luminance_target = 128;break;
    default: return 1;
  }

  luminance_avg = pixel_avg(q, scan);
  luminance_dif = luminance_target - luminance_avg;
  brightness_cur = qc_getbrightness(q);

#ifdef DEBUG
  fprintf(stderr, "Brightness: %d  Pixel luminance avg/target: %d/%d\n",
	  brightness_cur,luminance_avg, luminance_target );
#endif

  if (luminance_dif == 0) {
    return 0;
  } else if (luminance_dif > 0) {
    brightness_adj = luminance_dif / 2 + 1;
  } else {
    brightness_adj = luminance_dif / 2 - 1;
  }

  brightness_new = brightness_cur + brightness_adj;

  /* Adjusted brightness is out of range ..
   * throw in the towel ... auto-exposure has failed!
   */
  if (qc_setbrightness(q, brightness_new)) {
    fprintf(stderr, "Autoexposure failed!\n");
    return 0;
  }

  return 1;
}

int main(int argc, char **argv)
{
  scanbuf *scan = NULL;
  int arg;
  extern char *optarg;
  extern int optind;
  struct qcam *q;
  int width,height;
  int autocalibrate = 0;
  int autoexposure = 0;
  int snapshots = 0;
  int maxsnaps = 11;
  static char *getargs = "hx:y:p:b:d:H:S:B:c:w:s:t:l:f:EVWu";
  int found;
#ifdef DEBUG
  struct timeval start, end;
  double microseconds;
#endif

  if(geteuid()) {
    fprintf(stderr,"%s: Not installed SUID or run as root.  Exiting.\n",
	    argv[0]);
    exit(1);
  }

  q=qc_init();

  /* unpleasant pre-search for -f option */
  found = 0;
  while ((arg = getopt(argc, argv, getargs)) > 0) {
    if (arg == 'f' && !found) {
      qc_initfile(q, optarg);
      found = 1;
    }
  }
  if (!found) qc_initfile(q, 0); /* use default */

  qc_getresolution(q,&width,&height);

  /* Read command line */

  optind = 0;
  while((arg=getopt(argc,argv,getargs))>0) { 
    switch (arg) {
    case 'x':
      width=atoi(optarg);
      break;
    case 'y':
      height=atoi(optarg);
      break;
    case 'p':
      if (!getuid())
	q->port=strtol(optarg,NULL,0);
      break;
    case 'B':
      qc_setbitdepth(q,atoi(optarg));
      break;
    case 'b':
      qc_setbrightness(q,atoi(optarg));
      break;
    case 'H':
      qc_sethue(q,atoi(optarg));
      break;
    case 'S':
      qc_setsaturation(q,atoi(optarg));
      break;
    case 'c':
      qc_setcontrast(q,atoi(optarg));
      break;
    case 'w':
      qc_setwhitebal(q,atoi(optarg));
      break;
    case 'd':
      qc_setblacklevel(q,atoi(optarg));
      break;
    case 's':
      if (qc_settransfer_scale(q, atoi(optarg)))
	fprintf(stderr, "Bad scaling factor, valid values are 1, 2 or 4\n");
      break;
    case 't':
      if (qc_settop(q, atoi(optarg)))
	fprintf(stderr, "Bad top line, valid values are 1 - 243\n");
      break;
    case 'l':
      if (qc_setleft(q, atoi(optarg)))
	fprintf(stderr, "Bad left column, valid values are 2 - 336, and must be even\n");
      break;
    case 'f':
      /* already handled */
      break;
    case 'V':
      fprintf(stderr,"%s: Version 0.6\n",argv[0]);
      exit(0);
      break;
    case 'W':
      autocalibrate = 1;
      break;
    case 'E':
      autoexposure = 1;
      break;
    case 'u':
      qc_forceunidir(q);
      break;
    case 'h':
      usage();
      exit(0);
      break;
    default:
      fprintf(stderr,"%s: Unknown option or error in option\n",argv[0]);
      usage();
      exit(1);
      break;
    }
  }

  if(qc_setresolution(q,width,height)) {
    fprintf(stderr,"Invalid resolution.  Exiting.\n");
    exit(1);
  }

  /* Attempt to get permission to access IO ports.  Must be root */

  if (qc_open(q)) {
    fprintf(stderr,"Cannot open QuickCam; exiting.\n");
    exit(1);
  }

  setuid(getuid());

  /* ------------------------------------------------------------------ */
  do {

    /* Program the QuickCam */

    qc_set(q);

    if (autocalibrate && q->cam_version == 0x01) {
      qc_calibrate(q);
    }

    /* Scan one image */

#ifdef DEBUG
    if (gettimeofday(&start, NULL))
      perror("get start time");
#endif

    /* free(scan); */
    scan=qc_scan(q);

#ifdef DEBUG
    if (gettimeofday(&end, NULL))
      perror("get end time");

    microseconds = end.tv_usec - start.tv_usec;
    microseconds += 1000000.0 * (end.tv_sec - start.tv_sec);  
    fprintf(stderr, "%.0f ms to scan image", microseconds/1000.0);
    if (microseconds != 0)
      fprintf(stderr, " (%.1f fps)\n", 1000000.0/microseconds);
    else
      fprintf(stderr, "\n");
#endif

    snapshots++;
  } while (autoexposure && adjust_exposure(q, scan) && snapshots < maxsnaps);
  /* ------------------------------------------------------------------ */

  /* Output the image to stdout */

  if (q->cam_version == 0x10) qc_writeppm(q, stdout, scan);
                         else qc_writepgm(q,stdout,scan);
  free(scan);

  /* Release IO privileges */
  qc_close(q);

  return 0;
}

